/**
 * oddness_2factor_slower.c
 * 
 * Oddness_2factor_slower: A program which computes the oddness of a cubic bridgeless graph
 * (The oddness of a bridgeless cubic graph is the minimum number of odd components in 
 * any 2-factor of the graph.)
 *
 * Author: Jan Goedgebeur (jan.goedgebeur@ugent.be)
 *
 * ----------------------------------------
 * 
 * 
 * Copyright (c) 2016-2018 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * Compile with:
 *
 * cc -O4 oddness_2factor_slower.c -o oddness_2factor_slower
 *
 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>
#include "util.h"


/*************************Defines and global variables*************************/


/* Some marks */
unsigned int marks[MAXN];

#define MAXVAL INT_MAX - 1
static int markvalue = MAXVAL;
#define RESETMARKS {int mki; if ((markvalue += 1) > MAXVAL) \
      { markvalue = 1; for (mki=0;mki<MAXN;++mki) marks[mki]=0;}}
#define MARK(v) marks[v] = markvalue
#define UNMARK(v) marks[v] = markvalue - 1
#define ISMARKED(v) (marks[v] == markvalue)

int number_of_marked_vertices = 0;


unsigned char current_cycle[MAXN];
static int current_cycle_size = 0;


static int num_cycles_in_2_factor = 0;
static int num_odd_cycles_in_2_factor = 0;

static int min_num_odd_cycles_in_2_factor = 0;

static unsigned long long int all_2_factors_only_have_odd_cycles = 0;

static unsigned long long int frequency_min_odd_components[MAXN + 1] = {0};

static unsigned long long int times_cycle_is_canonical = 0;
static unsigned long long int times_cycle_is_not_canonical = 0;

/*********************************Methods**************************************/

int oddness_two_reached = 0;


/**************************************************************************/

/**
 * Tests if there is a vertex which is not part of a cycle (i.e. unmarked) but where
 * at least two of its neighbours are already part of a cycle. 
 * If this is the case, this current set of cycles cannot be extended to a 2-factor!
 * @return 
 */
int contains_unmarked_vertex_with_at_least_two_marked_neighbours(GRAPH graph) {
    int i, j;
    for(i = 0; i < number_of_vertices; i++)
        if(!ISMARKED(i)) {
            int num_marked_nbrs = 0;
            for(j = 0; j < REG; j++)
                if(ISMARKED(graph[i][j]))
                    num_marked_nbrs++;
            if(num_marked_nbrs >= 2)
                return 1;
        }
    
    return 0;
}

/**************************************************************************/

void search_cycle(GRAPH graph, unsigned char current_vertex, unsigned char previous_vertex,
        unsigned char first_vertex_of_cycle, unsigned char second_vertex_of_cycle, int cycle_length) {

    if(oddness_two_reached)
        return;


    if(current_vertex == first_vertex_of_cycle) { //I.e. cycle was formed

        //Test if there is a vertex which is not part of a cycle yet but where 
        //at least two of its neighbours are already part of a cycle.
        //BC helps a lot!
        if(contains_unmarked_vertex_with_at_least_two_marked_neighbours(graph)) {
            //Cannot lead to 2-factor so bound
            return;
        }

        num_odd_cycles_in_2_factor += ((cycle_length % 2) == 1);

        if(number_of_marked_vertices != number_of_vertices) {
            if(min_num_odd_cycles_in_2_factor == -1 || num_odd_cycles_in_2_factor < min_num_odd_cycles_in_2_factor) {
                //The above bounding crit doesn't really help

                //i.e. still vertices which weren't included in a cycle yet

                //Backup current_cycle[]
                int current_cycle_size_backup = current_cycle_size;
                unsigned char current_cycle_backup[current_cycle_size];
                memcpy(current_cycle_backup, current_cycle, sizeof (unsigned char) * current_cycle_size);

                int i, j, k;
                //unsigned long long int new_vertex_cycle_bitvector;
                /*
                 * To make sure that one won't find the same 2-factors multiple times,
                 * first vertex of new cycle must be > first vertex of previous cycle.
                 */
                for(i = second_vertex_of_cycle + 1; i < number_of_vertices; i++) {
                    //if((forbidden_vertices & BIT(i)) == 0) {
                    if(!ISMARKED(i)) { //Now start from next unmarked
                        for(j = 0; j < REG; j++) {
                            //If i > graph[i][j], it cannot be in canonical form!
                            //if(i < graph[i][j] && (forbidden_vertices & BIT(graph[i][j])) == 0) {
                            if(i < graph[i][j] && !ISMARKED(graph[i][j])) {
                                for(k = j + 1; k < REG; k++)
                                    if(i < graph[i][k] && graph[i][j] < graph[i][k] && !ISMARKED(graph[i][k])) {

                                        MARK(graph[i][j]);
                                        number_of_marked_vertices++;
                                        current_cycle_size = 0;
                                        current_cycle[current_cycle_size++] = graph[i][j];

                                        MARK(i);
                                        number_of_marked_vertices++;
                                        current_cycle[current_cycle_size++] = i;

                                        //Search next cycle
                                        //search_cycle(graph, graph[i][j], i, i, 1);
                                        search_cycle(graph, graph[i][k], i, graph[i][j], i, current_cycle_size);

                                        UNMARK(i);
                                        number_of_marked_vertices--;

                                        UNMARK(graph[i][j]);
                                        number_of_marked_vertices--;

                                    }

                            }
                        }
                        break; //Should only start from next unmarked!
                    }
                }

                if(i == number_of_vertices) {
                    fprintf(stderr, "Error: all remaining vertices were already marked!\n");
                    exit(1);
                }

                //Restore current_cycle[]
                current_cycle_size = current_cycle_size_backup;
                memcpy(current_cycle, current_cycle_backup, sizeof (unsigned char) * current_cycle_size);

            }
        } else {
            //All vertices are contained a cycle, so 2-factor found..
            /*
                            fprintf(stderr, "2-factor found\n");
                            fprintf(stderr, "num_cycles_in_2_factor: %d\n", num_cycles_in_2_factor);
                            fprintf(stderr, "num_odd_cycles_in_2_factor: %d\n", num_odd_cycles_in_2_factor);
             */


            if(num_odd_cycles_in_2_factor < min_num_odd_cycles_in_2_factor
                    || min_num_odd_cycles_in_2_factor == -1)
                min_num_odd_cycles_in_2_factor = num_odd_cycles_in_2_factor;

            if(num_odd_cycles_in_2_factor == 2)
                oddness_two_reached = 1;

        }
        //forbidden_vertices &= ~(*current_vertex_cycle_bitvector);

        num_cycles_in_2_factor--;
        num_odd_cycles_in_2_factor -= ((cycle_length % 2) == 1);
        //} else if((BIT(current_vertex) & *current_vertex_cycle_bitvector) > 0) {
        /*
            } else if(ISMARKED(current_vertex)) {
        
                fprintf(stderr, "Vertex was already marked. Cannot happen?\n");
                exit(1);
                //I.e. vertex was already on cycle (and is different from first_cycle_vertex)
                return;
         */
    } else {
        //Mark
        //*current_vertex_cycle_bitvector |= BIT(current_vertex);
        MARK(current_vertex);
        number_of_marked_vertices++;
        current_cycle[current_cycle_size++] = current_vertex;

        //I.e. vertex not visited before
        int i;
        for(i = 0; i < REG; i++) {
            //If graph[current_vertex][i] < current_cycle[0], the cycle cannot be canonical
            if(graph[current_vertex][i] >= second_vertex_of_cycle && graph[current_vertex][i] != previous_vertex
                    //&& (forbidden_vertices & BIT(graph[current_vertex][i])) == 0) {
                    //&& !ISMARKED(graph[current_vertex][i])) {
                    && (!ISMARKED(graph[current_vertex][i]) || graph[current_vertex][i] == first_vertex_of_cycle)) {
                search_cycle(graph, graph[current_vertex][i], current_vertex, first_vertex_of_cycle,
                        second_vertex_of_cycle, cycle_length + 1);
            }
        }

        //Unmark
        //*current_vertex_cycle_bitvector &= ~BIT(current_vertex);
        UNMARK(current_vertex);
        number_of_marked_vertices--;
        current_cycle_size--;
    }

}

/**************************************************************************/

int has_oddness_more_than_two(GRAPH graph) {
    //forbidden_vertices = 0;
    num_cycles_in_2_factor = num_odd_cycles_in_2_factor = 0;

    min_num_odd_cycles_in_2_factor = -1;

    oddness_two_reached = 0;
    
    //For all edges of the graph
    //Make all cycles where that edge is part of
    int i, j, k;
    //for(i = 0; i < number_of_vertices; i++) {
    i = 0; //Only need to start from 0 since else this vertex cannot be reached anymore!
        for(j = 0; j < REG; j++) {
            //If i > graph[i][j], it cannot be in canonical form!
            if(i < graph[i][j]) {
                for(k = j + 1; k < REG; k++)
                    if(i < graph[i][k] && graph[i][j] < graph[i][k]) {
                        //current_vertex_cycle_bitvector = BIT(i);
                        RESETMARKS;
                        number_of_marked_vertices = 0;
                        
                        MARK(graph[i][j]);
                        number_of_marked_vertices++;
                        current_cycle_size = 0;
                        current_cycle[current_cycle_size++] = graph[i][j];                        

                        MARK(i);
                        number_of_marked_vertices++;
                        current_cycle[current_cycle_size++] = i;

                        //Helps. About 40% faster...
                        
                        //search_cycle(graph, graph[i][j], i, i, 1);
                        search_cycle(graph, graph[i][k], i, graph[i][j], i, current_cycle_size);

                        if(oddness_two_reached) {
                            frequency_min_odd_components[min_num_odd_cycles_in_2_factor]++;
                            return 0;
                        }
                    }
            }
        }
    //}
    
    if(min_num_odd_cycles_in_2_factor == -1) {
        fprintf(stderr, "Error: no 2-factor found!\n");
        printgraph(graph);
        exit(1);
    }
    
    frequency_min_odd_components[min_num_odd_cycles_in_2_factor]++;
    
    return 1;


}

/**************************************************************************/

void check_graph(GRAPH graph) {
    int i, j, k;
    for(i = 0; i < number_of_vertices; i++)
        for(j = 0; j < REG; j++)
            for(k = j+1; k < REG; k++)
                if(graph[i][j] > graph[i][k]) {
                    fprintf(stderr, "Error: it is assumed that the neighbours of each vertex are listed in ascending order!\n");
                    exit(1);
                }
}

/**************************************************************************/

/*
 *
 */
int main(int argc, char** argv) {

    if(argc != 2) {
        fprintf(stderr, "Error: invalid number of arguments.\n");
        fprintf(stderr, "Usage: %s <number_of_vertices>\n", argv[0]);
        exit(1);
    } else {
        number_of_vertices = atoi(argv[1]);
        if(number_of_vertices > MAXN) {
            fprintf(stderr, "Number of vertices is too big (limit is %d) \n", MAXN);
            exit(0);
        }
    }
    
    
    fprintf(stderr, "Warning: inputgraphs are assumed to be bridgeless cubic graphs!\n");
    
    fprintf(stderr, "Warning: this version assumes that the graphs are class 2 (so have oddness at least 2)\n");
    
    //For the timing
    struct tms TMS;        

    int codelength = number_of_vertices * REG / 2 + number_of_vertices;
    unsigned char code[codelength];
    unsigned long long int number_of_graphs_read = 0;
    unsigned long long int number_of_graphs_oddness_more_than_two = 0;

    while(fread(code, sizeof(unsigned char), codelength, stdin)) {
        number_of_graphs_read++;

        decode_multicode(code, codelength);
        check_graph(graph); //Needed now!
        //printgraph(graph);

        if(has_oddness_more_than_two(graph)) {
            if(fwrite(code, sizeof(unsigned char), codelength, stdout) < codelength) {
                fprintf(stderr, "Error: couldn't write code!\n");
                exit(1);
            }
            number_of_graphs_oddness_more_than_two++;
        }

    }

    fprintf(stderr, "Tested %lld graphs\n", number_of_graphs_read);
    fprintf(stderr, "Found %lld graphs which have oddness at least 4\n", number_of_graphs_oddness_more_than_two);

    //Temp
/*
    fprintf(stderr, "Times cycle is canonical: %llu (%.1f%%)\n", times_cycle_is_canonical,
            (double) 100 * times_cycle_is_canonical / (times_cycle_is_canonical + times_cycle_is_not_canonical));
*/

    int i;
    for(i = 0; i <= MAXN; i++)
        if(frequency_min_odd_components[i] > 0)
            fprintf(stderr, "Num graphs with min %d odd components: %llu\n", i, frequency_min_odd_components[i]);
    
    times(&TMS);
    unsigned int savetime = (unsigned int) TMS.tms_utime;    

    fprintf(stderr, "CPU time filter: %.1f seconds.\n", (double) savetime / (double) CLK_TCK);        
    
    return(EXIT_SUCCESS);
}