/**
 * oddness_perf_matching.c
 * 
 * Oddness_perf_matching: A program which computes the oddness of a cubic bridgeless graph
 * (The oddness of a bridgeless cubic graph is the minimum number of odd components in 
 * any 2-factor of the graph.)
 *
 * Author: Jan Goedgebeur (jan.goedgebeur@ugent.be)
 *
 * ----------------------------------------
 * 
 * 
 * Copyright (c) 2016-2018 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * Compile with:
 *
 * cc -O4 oddness_perf_matching.c -o oddness_perf_matching
 *
 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>
#include "util.h"

/*************************Defines and global variables*************************/

/* Some marks */
unsigned int marks[MAXN];

#define MAXVAL INT_MAX - 1
static int markvalue = MAXVAL;
#define RESETMARKS {int mki; if ((markvalue += 1) > MAXVAL) \
      { markvalue = 1; for (mki=0;mki<MAXN;++mki) marks[mki]=0;}}
#define MARK(v) marks[v] = markvalue
#define UNMARK(v) marks[v] = markvalue - 1
#define ISMARKED(v) (marks[v] == markvalue)


unsigned int marks2[MAXN];

#define MAXVAL INT_MAX - 1
static int markvalue2 = MAXVAL;
#define RESETMARKS2 {int mki; if ((markvalue2 += 1) > MAXVAL) \
      { markvalue2 = 1; for (mki=0;mki<MAXN;++mki) marks2[mki]=0;}}
#define MARK2(v) marks2[v] = markvalue2
#define UNMARK2(v) marks2[v] = markvalue2 - 1
#define ISMARKED2(v) (marks2[v] == markvalue2)

int oddness_0_found = 0;
int parity_previous_2factor = -1;

int markedneighbour[MAXN];

static unsigned long long int number_of_perfect_matchings = 0;

int min_num_odd_cycles;
static unsigned long long int frequency_min_odd_components[MAXN+1] = {0};

/*********************************Methods**************************************/

void print_2factor_sizes() {
    RESETMARKS2;
    int i;
    int total_size = 0;
    for(i = 0; i < number_of_vertices && total_size < number_of_vertices; i++)
        if(!ISMARKED2(i)) {
            int next_vertex = i;
            int j = 0;
            int size = 0;
            fprintf(stderr, "( ");
            while(j < REG) {
                //fprintf(stderr, "%d ", next_vertex);
                MARK2(next_vertex);
                size++;
                for(j = 0; j < REG; j++)
                    if(!ISMARKED2(graph[next_vertex][j]) && graph[next_vertex][j] != markedneighbour[next_vertex]) {
                        next_vertex = graph[next_vertex][j];
                        break;
                    }
            }
            fprintf(stderr, "%d ", size);
            fprintf(stderr, ") ");
            total_size += size;
        }
    fprintf(stderr, "\n");
}

/******************************************************************************/

void print_2factor() {
    fprintf(stderr, "%llu: ", number_of_perfect_matchings);
    RESETMARKS2;
    int i;
    int total_size = 0;
    for(i = 0; i < number_of_vertices && total_size < number_of_vertices; i++)
        if(!ISMARKED2(i)) {
            int next_vertex = i;
            int j = 0;
            fprintf(stderr, "( ");
            while(j < REG) {
                fprintf(stderr, "%d ", next_vertex);
                MARK2(next_vertex);
                for(j = 0; j < REG; j++)
                    if(!ISMARKED2(graph[next_vertex][j]) && graph[next_vertex][j] != markedneighbour[next_vertex]) {
                        next_vertex = graph[next_vertex][j];
                        break;
                    }
            }
            fprintf(stderr, ") ");
        }
    fprintf(stderr, "\n");
}

/******************************************************************************/

void print_perfect_matching() {
    fprintf(stderr, "%llu: ", number_of_perfect_matchings);
    int i;
    for(i = 0; i < number_of_vertices; i++)
        if(i < markedneighbour[i])
            fprintf(stderr, "(%d,%d) ", i, markedneighbour[i]);
    fprintf(stderr, "\n");
}

/******************************************************************************/

void count_num_odd_cycles() {
    RESETMARKS2;
    int i;
    int total_size = 0;
    int num_odd_cycles = 0;
    for(i = 0; i < number_of_vertices && total_size < number_of_vertices; i++)
        if(!ISMARKED2(i)) {
            int next_vertex = i;
            int j = 0;
            int size = 0;
            //fprintf(stderr, "( ");
            while(j < REG) {
                //fprintf(stderr, "%d ", next_vertex);
                MARK2(next_vertex);
                size++;
                for(j = 0; j < REG; j++)
                    if(!ISMARKED2(graph[next_vertex][j]) && graph[next_vertex][j] != markedneighbour[next_vertex]) {
                        next_vertex = graph[next_vertex][j];
                        break;
                    }
            }
            //fprintf(stderr, "%d ", size);
            //fprintf(stderr, ") ");
            total_size += size;
            
            if((size % 2) == 1)
                num_odd_cycles++;
        }
    //fprintf(stderr, "\n");
    
    //fprintf(stderr, "Num odd cycles: %d\n", num_odd_cycles);
    if(num_odd_cycles < min_num_odd_cycles)
        min_num_odd_cycles = num_odd_cycles;
    
    if(num_odd_cycles == 0) //Oddness can't be smaller than 0, so we can stop
        oddness_0_found = 1;
}

/******************************************************************************/

void construct_perfect_matching() {
    if(oddness_0_found)
       return;

    int i;
    for(i = 0; i < number_of_vertices; i++)
        if(!ISMARKED(i))
            break;
    
    if(i == number_of_vertices) {
        //Perfect matching found
        number_of_perfect_matchings++;
        count_num_odd_cycles();
    } else {
        MARK(i);
        int j;
        for(j = 0; j < REG; j++)
            if(!ISMARKED(graph[i][j])) {
                MARK(graph[i][j]);
                markedneighbour[i] = graph[i][j];
                markedneighbour[graph[i][j]] = i;

                //recursion
                construct_perfect_matching();

                UNMARK(graph[i][j]);

            }

        UNMARK(i);
    }
}

/******************************************************************************/

/**
 * Returns 1 if the intputgraph has odness larger than 2, else returns 0.
 */
int has_oddness_larger_than_two() {
    RESETMARKS;
    
    min_num_odd_cycles = MAXN;

    number_of_perfect_matchings = 0;
    parity_previous_2factor = -1;
    oddness_0_found = 0;
    
    //Vertices which are already part of the perfect matching which is being constructed are marked
    MARK(0);
    int i = 0;
    for(i = 0; i < REG; i++) {
        MARK(graph[0][i]);
        markedneighbour[0] = graph[0][i];
        markedneighbour[graph[0][i]] = 0;

        construct_perfect_matching();

        UNMARK(graph[0][i]);
    }


    //fprintf(stderr, "Number of perfect matchings: %lld\n", number_of_perfect_matchings);
    //fprintf(stderr, "Min num odd cycles (i.e. oddness): %d\n", min_num_odd_cycles);
    
    frequency_min_odd_components[min_num_odd_cycles]++;

    return min_num_odd_cycles > 2;

}

/******************************************************************************/

/**
 * Code the cubic graph using the multicode format.
 */
void code_multicode(unsigned char code[], GRAPH g, int num_of_vertices) {
    int i, j;
    int codelength = 0;
    code[codelength++] = num_of_vertices;
    for(i = 0; i < num_of_vertices - 1; i++) {
        for(j = 0; j < REG; j++)
            if(g[i][j] > i)
                code[codelength++] = g[i][j] + 1;
        code[codelength++] = 0;
    }
    
    fprintf(stderr, "codelength: %d\n", codelength);

    //return codelength;
}

/******************************************************************************/

int is_neighbour(GRAPH graph, int v1, int v2) {
    int i;
    for(i = 0; i < REG; i++)
        if(graph[v1][i] == v2)
            return 1;
    return 0;
}

/******************************************************************************/

/*
 *
 */
int main(int argc, char** argv) {

    if(argc != 2) {
        fprintf(stderr, "Error: invalid number of arguments.\n");
        fprintf(stderr, "Usage: %s <number_of_vertices>\n", argv[0]);
        exit(1);
    } else {
        number_of_vertices = atoi(argv[1]);
        if(number_of_vertices > MAXN) {
            fprintf(stderr, "Number of vertices is too big (limit is %d) \n", MAXN);
            exit(0);
        }
    }

    //fprintf(stderr, "REG: %d and MAXN: %d\n", REG, MAXN);
    fprintf(stderr, "Warning: inputgraphs are assumed to be bridgeless cubic graphs!\n");
    

    //For the timing
    struct tms TMS;

    int codelength = number_of_vertices * REG / 2 + number_of_vertices;
    unsigned char code[codelength];
    unsigned long long int number_of_graphs_read = 0;
    unsigned long long int num_graphs_oddness_larger_than_two = 0;
    
    while(fread(code, sizeof (unsigned char), codelength, stdin)) {
        number_of_graphs_read++;

        decode_multicode(code, codelength);
        
        //printgraph(graph);
        
        if(has_oddness_larger_than_two()) {
            if(fwrite(code, sizeof (unsigned char), codelength, stdout) < codelength) {
                fprintf(stderr, "Error: couldn't write code!\n");
                exit(1);
            }
            num_graphs_oddness_larger_than_two++;
        }
    }

    times(&TMS);
    unsigned int savetime = (unsigned int) TMS.tms_utime;
    
    int i;
    for(i = 0; i <= MAXN; i++)
        if(frequency_min_odd_components[i] > 0)
            fprintf(stderr, "Num graphs with min %d odd components: %llu\n", i, frequency_min_odd_components[i]);    

    fprintf(stderr, "Read %lld graphs\n", number_of_graphs_read);
    fprintf(stderr, "Found %lld graphs with oddness at least 4\n", num_graphs_oddness_larger_than_two);
    fprintf(stderr, "CPU time filter: %.1f seconds.\n", (double) savetime / (double) CLK_TCK);

    return (EXIT_SUCCESS);
}
